#include <avr/power.h>

// pinout
const int LED1 = 5; // (digital) for debugging
const int LED2 = 6; // (digital) for debugging
const int LED3 = 2; // (digital) for debugging
const int SENSORPIN_DIGITAL = 9; // (pwm) frequency generator for capacitive sensor
const int SENSORPIN_ANALOG = 0;  // (analog) outputs capacitive sensor value
const int MOTOR_EN = 4;          // (digital) enable the driver to be active (HIGH) or sleep-mode (LOW)
const int MOTOR_PIN1 = 8;        // (digital) direction 1 - up
const int MOTOR_PIN2 = 7;        // (digital) direction 2 - down
const int MOTOR_PWMPIN = 3;      // (pwm) outputs speed for motor 
const int MOTOR_FEEDBACKPIN = 1; // (analog) feedback pin for motor driver : outputs motor current consumption   `
const int STRETCH_PIN = 2;       // (analog) outputs motor position (actually returns stretch sensor value)

// capacitive value reading related
const int STEPS = 128;        // for frequency generation
float values[STEPS];
float alpha = 0.01;           // value between 0 and 1 - the higher it is, the smoother (but slower) it gets 
int sensorValue;              // capacitive touch output value
const int TRIGGERVALUE = 600; // under what sensor value should sensor pin be triggered ?

// click related
bool buttonPin;                                 // simulate a pushbutton
bool one_click;                                 // determine if one click happened before a hold event - thus should the motor goes up or down
unsigned int one_click_timer, hold_click_timer; // those are used in order to define the maximum time a short click is memorized
unsigned int MAX_ClICK_TIMER = 3000;            // if a short click happens but no long clicks follow in the next MAX_ClICK_TIMER seconds, then one_click reinitialized

// motor related
const int MOTOR_SPEED = 255; // motor speed
const int upper_pos = 184;   // upper position value read by stretch sensor
const int lower_pos = 94;    // lower position value read by stretch sensor

// mode ?
bool debug_mode = false; // LEDs and serial communication enabled
bool demo_mode = false;  // only LEDs enabled

//=================================================
// General setup

void setup (){
  if (debug_mode) Serial.begin(115200); // start Serial communication
  if (debug_mode || demo_mode){
    pinMode(LED1, OUTPUT);
    pinMode(LED2, OUTPUT);
    pinMode(LED3, OUTPUT);
  }

  // capacitive sensor related
  pinMode(SENSORPIN_ANALOG, INPUT);
  pinMode(SENSORPIN_DIGITAL, OUTPUT);
  // initialize Timer1 with 0 (as we don't know the initial values)
  TCCR1A = 0; // set Timer TCCR1A register to 0
  TCCR1B = 0; // set Timer TCCR1B register to 0
  TCCR1A |= (1 << COM1A0); // Toggle OC1A on Compare Match.
  TCCR1B |= (1 << WGM12);  // CTC mode

  // motor related
  pinMode(MOTOR_EN, OUTPUT);
  pinMode(MOTOR_PIN1, OUTPUT);
  pinMode(MOTOR_PIN2, OUTPUT);
  pinMode(MOTOR_PWMPIN, OUTPUT);
  digitalWrite(MOTOR_EN, LOW); // let the motor driver sleeps until touch is detected

  // pseudo power management (couldn't find better than that - it changes absolutely nothing)
  power_twi_disable(); // disable TWI (I2C)
  power_spi_disable(); // disable SPI
  bitSet (DIDR0, ADC0D);  // disable digital buffer on A0
  bitSet (DIDR0, ADC1D);  // disable digital buffer on A1
  bitSet (DIDR0, ADC2D);  // disable digital buffer on A2
  bitSet (DIDR0, ADC3D);  // disable digital buffer on A3
  bitSet (DIDR0, ADC4D);  // disable digital buffer on A4
  bitSet (DIDR0, ADC5D);  // disable digital buffer on A5
}

//=================================================
// Loop

void loop (){
  getSensorValue();               // pretty easy to understand
  if (debug_mode) getPosition();  // get motor position thx to stretch pin
  int b = checkButton();          // check whether a click, a double-click, hold event or long hold event is happening
  switch (b){                     // actually here we only care about clicks and hold events
    case 1:
      clickEvent();
      break;
//    case 2:
//      doubleClickEvent();
//      break;
    case 3:
      holdEvent();
      break;
//    case 4:
//      longHoldEvent();
//      break;
  }
//  delay(1); // for stability
}

//=================================================
// Motor state

// motor current consumption
void getMotorConsumption(){
  // MC33926 datasheet : "The current sense output is approximately 525 mV/A"
  Serial.print("Motor consumption : ");
  Serial.print(analogRead(MOTOR_FEEDBACKPIN) * 9); // 5V / 1024 ADC counts / 525 mV per A = 9 mA per count
  Serial.println(" mA");
}

// motor position
void getPosition(){
  Serial.print("Motor position : ");
  Serial.println(analogRead(STRETCH_PIN));
}

//=================================================
// Capacitive sensor

void getSensorValue(){
  sensorValue = 0;

  for (int i = 0; i < STEPS; i++) {
    TCCR1B &= 0xFE; // turns off timer
    TCNT1 = 0;      // resets timer counter register
    OCR1A = i;      // sets new frequency step
    TCCR1B |= 0x01; // turns on timer

    values[i] = values[i] * alpha + analogRead(SENSORPIN_ANALOG) * (1 - alpha); // exponential moving avg
    
    if (values[i] > sensorValue) { // finds the signal peak
      sensorValue = values[i];
    }
  }

  if (debug_mode){ // check on the Serialmonitor your values
    Serial.print("Capacitive output : ");
    Serial.println(sensorValue);
  }
  
  if (sensorValue < TRIGGERVALUE){
    if (debug_mode || demo_mode){
      digitalWrite(LED1, HIGH);
      digitalWrite(LED2, LOW);
    }
    buttonPin = false;
  } else {
    if (debug_mode || demo_mode){
      digitalWrite(LED1, LOW);
      digitalWrite(LED2, HIGH);
    }
    buttonPin = true;
  }

  delay(50); // don't run too fast for the sensor not to be too sensitive
}

//=================================================
// Motor possible motions

// going up
void motionUp(){
  digitalWrite(MOTOR_EN, HIGH); // activate motor driver
  while (buttonPin == false){// && analogRead(STRETCH_PIN) < upper_pos){
    getSensorValue();
    if (debug_mode || demo_mode) digitalWrite(LED3, HIGH);
    digitalWrite(MOTOR_PIN2, LOW);
    digitalWrite(MOTOR_PIN1, HIGH);
    analogWrite(MOTOR_PWMPIN, MOTOR_SPEED);
    if (debug_mode){
      getMotorConsumption();
      getPosition();
    }
  }
  if (debug_mode || demo_mode) digitalWrite(LED3, LOW);
  digitalWrite(MOTOR_PIN1, LOW);
  digitalWrite(MOTOR_PIN2, LOW);
  digitalWrite(MOTOR_EN, LOW);
}

// going down
void motionDown(){
  digitalWrite(MOTOR_EN, HIGH); // activate motor driver
  while (buttonPin == false){ //&& analogRead(STRETCH_PIN) > lower_pos){
    getSensorValue();
    if (debug_mode || demo_mode) digitalWrite(LED3, HIGH);
    digitalWrite(MOTOR_PIN1, LOW);
    digitalWrite(MOTOR_PIN2, HIGH);
    analogWrite(MOTOR_PWMPIN, MOTOR_SPEED);
    if (debug_mode){
      getMotorConsumption();
      getPosition();
    }
  }
  if (debug_mode || demo_mode) digitalWrite(LED3, LOW);
  digitalWrite(MOTOR_PIN1, LOW);
  digitalWrite(MOTOR_PIN2, LOW);
  digitalWrite(MOTOR_EN, LOW);
}

//=================================================
//  MULTI-CLICK:  One Button, Multiple Events

// a short click is detected
void clickEvent() {
  one_click = true;           // one click
  one_click_timer = millis(); // 1st time capture
}

// a long click is detected
void holdEvent() {
  hold_click_timer = millis(); // 2nd time capture for comparison
  
  // if a short click is detected in the previous "MAX_ClICK_TIMER" seconds
  // goes down
  if (one_click && ((hold_click_timer - one_click_timer) <= MAX_ClICK_TIMER)) motionDown();
  // else goes up
  else motionUp();
  
  one_click = false; // reset the preparation for a hold event
}

// Button timing variables
int debounce = 10;          // ms, debounce period to prevent flickering when pressing or releasing the button
int DCgap = 50;             // ms, defines how short a short click is (also defines max ms between clicks for a double click event)
int holdTime = 550;         // ms, hold period: how long to wait for press+hold event
int longHoldTime = 3050;    // ms, long hold period: how long to wait for press+hold event

// Button variables
bool buttonVal = true;          // value read from button
bool buttonLast = true;         // buffered value of the button's previous state
bool DCwaiting = false;         // whether we're waiting for a double click (down)
bool DConUp = false;            // whether to register a double click on next release, or whether to wait and click
bool singleOK = true;           // whether it's OK to do a single click
long downTime = -1;             // time the button was pressed down
long upTime = -1;               // time the button was released
bool ignoreUp = false;          // whether to ignore the button release because the click+hold was triggered
bool waitForUp = false;         // when held, whether to wait for the up event
bool holdEventPast = false;     // whether or not the hold event happened already
bool longHoldEventPast = false; // whether or not the long hold event happened already

int checkButton(){   
   int event = 0;
   buttonVal = buttonPin; // here is to match our simulated pushbutton to the multi-click events
   
   // Button pressed down
   if (buttonVal == false && buttonLast == true && (millis() - upTime) > debounce)
   {
       downTime = millis();
       ignoreUp = false;
       waitForUp = false;
       singleOK = true;
       holdEventPast = false;
       longHoldEventPast = false;
       if ((millis()-upTime) < DCgap && DConUp == false && DCwaiting == true)  DConUp = true;
       else  DConUp = false;
       DCwaiting = false;
   }
   
   // Button released
   else if (buttonVal == true && buttonLast == false && (millis() - downTime) > debounce)
   {       
       if (not ignoreUp)
       {
           upTime = millis();
           if (DConUp == false) DCwaiting = true;
           else
           {
               event = 2;
               DConUp = false;
               DCwaiting = false;
               singleOK = false;
           }
       }
   }
   
   // Test for normal click event: DCgap expired
   if ( buttonVal == true && (millis()-upTime) >= DCgap && DCwaiting == true && DConUp == false && singleOK == true && event != 2)
   {
       event = 1;
       DCwaiting = false;
   }
   
   // Test for hold
   if (buttonVal == false && (millis() - downTime) >= holdTime) {
       // Trigger "normal" hold
       if (not holdEventPast)
       {
           event = 3;
           waitForUp = true;
           ignoreUp = true;
           DConUp = false;
           DCwaiting = false;
           //downTime = millis();
           holdEventPast = true;
       }
       
       // Trigger "long" hold
       if ((millis() - downTime) >= longHoldTime)
       {
           if (not longHoldEventPast)
           {
               event = 4;
               longHoldEventPast = true;
           }
       }
   }
   buttonLast = buttonVal;
   return event;
}
